using System;
using System.Collections.Generic;
using System.IO;
using System.Windows.Forms;
using Un4seen.Bass;
using Un4seen.Bass.AddOn.Mix;
using Un4seen.BassWasapi;

namespace BassMixer
{
  public partial class Main : Form
  {
    int BufferMs = 50;
    int songIndex = 0;
    string[] songs = null;
    int _mixer = 0;
    private int _syncProc = 0;

    private List<Stream> Streams = new List<Stream>(2);
    private int CurrentStreamIndex = 0;
    private SYNCPROC PlaybackEndProcDelegate = null;
    private WASAPIPROC _wasapiProc = null;
    private Timer _timer = null;

    public Main()
    {
      InitializeComponent();
    }

    /// <summary>
    /// Callback from WasApi to deliver data from Decoding channel
    /// </summary>
    /// <param name="buffer"></param>
    /// <param name="length"></param>
    /// <param name="user"></param>
    /// <returns></returns>
    private int WasApiCallback(IntPtr buffer, int length, IntPtr user)
    {
      if (_mixer == 0)
      {
        return 0;
      }
      try
      {
        return Bass.BASS_ChannelGetData(_mixer, buffer, length);
      }
      catch (Exception)
      { }
      return 0;
    }


    private void Main_Load(object sender, EventArgs e)
    {
      tbMusicDir.Text = @"";
      Bass.BASS_SetConfig(BASSConfig.BASS_CONFIG_BUFFER, BufferMs);
      if (!Bass.BASS_Init(0, 48000, 0, IntPtr.Zero))
      {
        MessageBox.Show(this, "Bass_Init error!");
      }

      _wasapiProc = new WASAPIPROC(WasApiCallback);

      BASS_WASAPI_DEVICEINFO[] deviceinfo = BassWasapi.BASS_WASAPI_GetDeviceInfos();

      for (int i = 0; i < deviceinfo.GetLength(0); i++)
      {
        if (!deviceinfo[i].IsInput)
        {
          comboBoxDevices.Items.Add(new Item(i, deviceinfo[i].name));
        }
      }

      _timer = new Timer();
      _timer.Interval = 1000;
      _timer.Tick += _timer_Tick;
    }

    private void Main_Close(object sender, FormClosingEventArgs e)
    {
      BassWasapi.BASS_WASAPI_Free();
      Bass.BASS_Stop();
      Bass.BASS_Free();
    }

    private void btSelectDir_Click(object sender, EventArgs e)
    {
      FolderBrowserDialog fbd = new FolderBrowserDialog();
      fbd.Description = "Select a Music Directory";
      if (fbd.ShowDialog() == DialogResult.OK)
      {
        tbMusicDir.Text = fbd.SelectedPath;
      }
    }

    private void btStart_Click(object sender, EventArgs e)
    {
      BASSWASAPIInit initFlags = BASSWASAPIInit.BASS_WASAPI_AUTOFORMAT | BASSWASAPIInit.BASS_WASAPI_EXCLUSIVE;
      if (BassWasapi.BASS_WASAPI_Init((comboBoxDevices.SelectedItem as Item).deviceNum, 44100, 2, initFlags, 0f, 0f, _wasapiProc, IntPtr.Zero))
      {
        for (int i = 0; i < 2; i++)
          Streams.Add(null);

        _mixer = BassMix.BASS_Mixer_StreamCreate(44100, 2, BASSFlag.BASS_MIXER_NONSTOP | BASSFlag.BASS_SAMPLE_FLOAT | BASSFlag.BASS_STREAM_DECODE);
        PlaybackEndProcDelegate = new SYNCPROC(PlaybackEndProc);
        int plugin = Bass.BASS_PluginLoad("bassflac.dll");

        BassWasapi.BASS_WASAPI_SetVolume(BASSWASAPIVolume.BASS_WASAPI_CURVE_DB, (float)100 / 100f);
        bool result = BassWasapi.BASS_WASAPI_Start();
      }
      else
        MessageBox.Show(this, "Bass_Init error!");

      songIndex = -1;
      songs = Directory.GetFiles(tbMusicDir.Text, "*.*");
      HandleNextSong();
      if (!BassWasapi.BASS_WASAPI_IsStarted())
      {
        BassWasapi.BASS_WASAPI_Start();
      }

      _timer.Start();
    }

    private void btStop_Click(object sender, EventArgs e)
    {
      BassWasapi.BASS_WASAPI_Stop(true);
      _timer.Stop();
    }

    private void btNext_Click(object sender, EventArgs e)
    {
      Stream stream = GetCurrentStream();
    }

    private void HandleNextSong()
    {
      songIndex++;

      if (songs.GetLength(0) <= songIndex)
      {
        return;
      }

      Stream stream = GetNextStream();

      int streamID = stream.ID;
      Streams[CurrentStreamIndex] = stream;

      if (streamID != 0)
      {
        AddText($"{songs[songIndex]} added to Mixer. Stream: {stream}");

        Bass.BASS_ChannelLock(_mixer, true);

        SetSyncProc(streamID, 0.0);
        BassMix.BASS_Mixer_StreamAddChannel(_mixer, streamID, BASSFlag.BASS_MIXER_NORAMPIN);

        Bass.BASS_ChannelLock(_mixer, false);
      }
      else
      {
        AddText($"Error: {Enum.GetName(typeof(BASSError), Bass.BASS_ErrorGetCode())}");
      }
    }

    private void SetSyncProc(int stream, double timePos)
    {
      long streamLen = Bass.BASS_ChannelGetLength(stream, BASSMode.BASS_POS_BYTES);
      double streamtime = Bass.BASS_ChannelBytes2Seconds(stream, streamLen);

      long mixerPos = Bass.BASS_ChannelGetPosition(_mixer, BASSMode.BASS_POS_BYTES | BASSMode.BASS_POS_DECODE);
      long streamBytes = Bass.BASS_ChannelSeconds2Bytes(_mixer, streamtime - timePos);
      long syncPos = mixerPos + streamBytes;

      BASS_CHANNELINFO streamInfo = Bass.BASS_ChannelGetInfo(stream);
      BASS_CHANNELINFO mixerInfo = Bass.BASS_ChannelGetInfo(_mixer);
      AddText($"Streamlen: {streamLen} StreamTime: {streamtime} MixerPos: {mixerPos} StreamBytes: {streamBytes} SyncPos: {syncPos}");
      AddText($"Stream: Channels: {streamInfo.chans} Rate: { streamInfo.freq} Flags: {streamInfo.flags}");
      AddText($"Mixer:  Channels: {mixerInfo.chans} Rate: {mixerInfo.freq} Flags: {mixerInfo.flags}");

      if (_syncProc != 0)
      {
        Bass.BASS_ChannelRemoveSync(_mixer, _syncProc);
      }

      _syncProc = Bass.BASS_ChannelSetSync(_mixer,
        BASSSync.BASS_SYNC_ONETIME | BASSSync.BASS_SYNC_POS | BASSSync.BASS_SYNC_MIXTIME,
        syncPos, PlaybackEndProcDelegate,
        IntPtr.Zero);

    }

    private void AddText(string text)
    {
      text += Environment.NewLine;
      textBoxStatus.AppendText(text);
    }

    private void PlaybackEndProc(int handle, int stream, int data, IntPtr userData)
    {
      AddText($"Playback ended: stream {stream}");
      HandleNextSong();
    }

    private void _timer_Tick(object sender, EventArgs e)
    {
      Stream stream = GetCurrentStream();
      int level = Bass.BASS_ChannelGetLevel(stream.ID);

      int peakL = Utils.LowWord32(level); // the left level
      int peakR = Utils.HighWord32(level); // the right level

      double dbLeft = Utils.LevelToDB(peakL, 32768);
      double dbRight = Utils.LevelToDB(peakR, 32768);

      AddText($"L: {dbLeft} R: {dbRight}");
    }

    private Stream GetCurrentStream()
    {
      if (CurrentStreamIndex < 0)
        CurrentStreamIndex = 0;
      else if (CurrentStreamIndex >= Streams.Count)
        CurrentStreamIndex = Streams.Count - 1;

      return Streams.Count > 0 ? Streams[CurrentStreamIndex] : null;
    }

    private Stream GetNextStream()
    {
      Stream currentStream = GetCurrentStream();

      Stream stream = new Stream(songs[songIndex]);

      if (currentStream == null)
        return stream;

      if (currentStream.ID != 0 || Bass.BASS_ChannelIsActive(currentStream.ID) != (int)BASSActive.BASS_ACTIVE_STOPPED)
      {
        CurrentStreamIndex++;

        if (CurrentStreamIndex >= Streams.Count)
          CurrentStreamIndex = 0;
      }
      return stream;
    }

  }

  public class Item
  {
    public int deviceNum;
    public string deviceName;

    public Item(int num, string name)
    {
      deviceName = name;
      deviceNum = num;
    }

    public override string ToString()
    {
      return deviceName;
    }
  }
}